@@ -27,8 +27,10 @@
 #include <gensio/gensio_class.h>
 #include <gensio/argvutils.h>
 #include <gensio/gensio_list.h>
+#include <gensio/gensio_ax25_addr.h>
 
 #include "errtrig.h"
+#include <pthread_handler.h>
 
 static const char *progname = "gensio";
 
@@ -61,6 +63,15 @@ const char *gensio_get_progname(void)
 #include <net/if.h>
 #include <limits.h>
 #include <dlfcn.h>
+#if USE_OPENPTY
+#include <util.h>
+#endif
+
+#if USE_GGL_INT
+#define GID_CAST (int *)
+#else
+#define GID_CAST
+#endif
 
 int
 gensio_unix_os_setupnewprog(void)
@@ -86,7 +97,7 @@ gensio_unix_os_setupnewprog(void)
     if (err)
 	return errno;
 
-    getgrouplist(pw->pw_name, pw->pw_gid, groups, &ngroup);
+    getgrouplist(pw->pw_name, pw->pw_gid, GID_CAST groups, &ngroup);
     if (ngroup > 0) {
 	int ngroup2 = ngroup;
 
@@ -94,7 +105,7 @@ gensio_unix_os_setupnewprog(void)
 	if (!groups)
 	    return ENOMEM;
 
-	err = getgrouplist(pw->pw_name, pw->pw_gid, groups, &ngroup2);
+	err = getgrouplist(pw->pw_name, pw->pw_gid, GID_CAST groups, &ngroup2);
 	if (err == -1) {
 	    err = errno;
 	    free(groups);
@@ -197,13 +208,29 @@ gensio_scan_network_port(struct gensio_os_funcs *o, const char *str,
 #endif
     }
 
-    if (strncmp(str, "unix,", 4) == 0 ||
-		(rargs && strncmp(str, "unix(", 4) == 0)) {
+    if (strncmp(str, "unix,", 5) == 0 ||
+		(rargs && strncmp(str, "unix(", 5) == 0)) {
 	if (family != AF_UNSPEC)
 	    return GE_INVAL;
 	str += 4;
     handle_unix:
 	protocol = GENSIO_NET_PROTOCOL_UNIX;
+#ifdef SOCK_SEQPACKET
+    } else if (strncmp(str, "unixseq,", 8) == 0 ||
+		(rargs && strncmp(str, "unixseq(", 8) == 0)) {
+	if (family != AF_UNSPEC)
+	    return GE_INVAL;
+	str += 7;
+    handle_unix_seqpacket:
+	protocol = GENSIO_NET_PROTOCOL_UNIX_SEQPACKET;
+#endif
+    } else if (strncmp(str, "unixdgram,", 10) == 0 ||
+		(rargs && strncmp(str, "unixdgram(", 10) == 0)) {
+	if (family != AF_UNSPEC)
+	    return GE_INVAL;
+	str += 9;
+    handle_unix_dgram:
+	protocol = GENSIO_NET_PROTOCOL_UNIX_DGRAM;
     } else if (strncmp(str, "tcp,", 4) == 0 ||
 		(rargs && strncmp(str, "tcp(", 4) == 0)) {
 	str += 3;
@@ -223,11 +250,17 @@ gensio_scan_network_port(struct gensio_os_funcs *o, const char *str,
 #else
 	return GE_NOTSUP;
 #endif
-    } else if (rprotocol && *rprotocol != 0) {
+    } else if (rprotocol && *rprotocol != GENSIO_NET_PROTOCOL_UNSPEC) {
 	doskip = false;
 	switch (*rprotocol) {
 	case GENSIO_NET_PROTOCOL_UNIX:
 	    goto handle_unix;
+	case GENSIO_NET_PROTOCOL_UNIX_DGRAM:
+	    goto handle_unix_dgram;
+#ifdef SOCK_SEQPACKET
+	case GENSIO_NET_PROTOCOL_UNIX_SEQPACKET:
+	    goto handle_unix_seqpacket;
+#endif
 	case GENSIO_NET_PROTOCOL_TCP:
 	    goto handle_tcp;
 	case GENSIO_NET_PROTOCOL_UDP:
@@ -291,10 +324,15 @@ gensio_os_scan_netaddr(struct gensio_os_funcs *o, const char *str, bool listen,
     struct gensio_addr *addr;
     int rv;
 
+    if (protocol == GENSIO_NET_PROTOCOL_UNSPEC && strncmp(str, "ax25:", 5) == 0)
+	return gensio_ax25_str_to_addr(o, str, raddr);
+
     rv = o->addr_scan_ips(o, str, listen, AF_UNSPEC,
 			  protocol, &is_port_set, true, &addr);
     if (!rv && !listen && !is_port_set &&
-		protocol != GENSIO_NET_PROTOCOL_UNIX) {
+		protocol != GENSIO_NET_PROTOCOL_UNIX &&
+		protocol != GENSIO_NET_PROTOCOL_UNIX_DGRAM &&
+		protocol != GENSIO_NET_PROTOCOL_UNIX_SEQPACKET) {
 	gensio_addr_free(addr);
 	rv = GE_INVAL;
     } else if (!rv) {
@@ -310,8 +348,11 @@ gensio_os_free_net_ifs(struct gensio_os_funcs *o,
 {
     unsigned int i, j;
 
+    if (!ifs)
+	return;
+
     for (i = 0; i < nifs; i++) {
-	if (!ifs)
+	if (!ifs[i])
 	    continue;
 	if (ifs[i]->name)
 	    gensio_os_funcs_zfree(o, ifs[i]->name);
@@ -724,7 +765,7 @@ struct gensio_win_commport
     BOOL orig_timeouts_set;
     COMMTIMEOUTS orig_timeouts;
 
-    BOOL hupcl; /* FIXME - implement this. */
+    BOOL hupcl;
     BOOL break_set;
     BOOL dtr_set;
     BOOL rts_set;
@@ -789,12 +830,12 @@ gensio_win_setup_commport(struct gensio_os_funcs* o, HANDLE h,
     if (!EscapeCommFunction(h, CLRBREAK))
 	goto out_err;
     c->break_set = FALSE;
-    if (!EscapeCommFunction(h, CLRRTS))
+    if (!EscapeCommFunction(h, SETRTS))
 	goto out_err;
-    c->rts_set = FALSE;
-    if (!EscapeCommFunction(h, CLRDTR))
+    c->rts_set = TRUE;
+    if (!EscapeCommFunction(h, SETDTR))
 	goto out_err;
-    c->dtr_set = FALSE;
+    c->dtr_set = TRUE;
 
     /* Break timer */
     c->break_timer = CreateWaitableTimer(NULL, FALSE, NULL);
@@ -833,6 +874,10 @@ gensio_win_cleanup_commport(struct gensio_os_funcs *o, HANDLE h,
 	SetCommState(h, &(*c)->orig_dcb);
     if ((*c)->orig_timeouts_set)
 	SetCommTimeouts(h, &(*c)->orig_timeouts);
+    if ((*c)->hupcl) {
+	EscapeCommFunction(h, CLRDTR);
+	EscapeCommFunction(h, CLRRTS);
+    }
     o->free(o, *c);
     *c = NULL;
 }
@@ -952,9 +997,9 @@ gensio_win_commport_control(struct gensio_os_funcs *o, int op, bool get,
 
     case GENSIO_IOD_CONTROL_LOCAL:
 	if (get)
-	    *((int *) val) = t->fDsrSensitivity;
+	    *((int *) val) = !t->fDsrSensitivity;
 	else
-	    t->fDsrSensitivity = val;
+	    t->fDsrSensitivity = !val;
 	break;
 
     case GENSIO_IOD_CONTROL_HANGUP_ON_DONE:
@@ -1053,13 +1098,13 @@ gensio_win_commport_control(struct gensio_os_funcs *o, int op, bool get,
 	if (!GetCommModemStatus(h, &dval))
 	    goto out_err;
 	if (dval & MS_CTS_ON)
-	    rval |= SERGENSIO_MODEMSTATE_CTS;
+	    rval |= GENSIO_SER_MODEMSTATE_CTS;
 	if (dval & MS_DSR_ON)
-	    rval |= SERGENSIO_MODEMSTATE_DSR;
+	    rval |= GENSIO_SER_MODEMSTATE_DSR;
 	if (dval & MS_RING_ON)
-	    rval |= SERGENSIO_MODEMSTATE_RI;
+	    rval |= GENSIO_SER_MODEMSTATE_RI;
 	if (dval & MS_RLSD_ON)
-	    rval |= SERGENSIO_MODEMSTATE_CD;
+	    rval |= GENSIO_SER_MODEMSTATE_CD;
 	*((int *) val) = rval;
 	break;
     }
@@ -1216,7 +1261,6 @@ gensio_win_do_exec(struct gensio_os_funcs *o,
 {
     int rv = 0;
     char *cmdline, *envb = NULL;
-    SECURITY_ATTRIBUTES sattr;
     STARTUPINFOA suinfo;
     PROCESS_INFORMATION procinfo;
     HANDLE stdin_m = NULL, stdin_s = NULL;
@@ -1238,22 +1282,19 @@ gensio_win_do_exec(struct gensio_os_funcs *o,
 	}
     }
 
-    memset(&sattr, 0, sizeof(sattr));
     memset(&suinfo, 0, sizeof(suinfo));
     memset(&procinfo, 0, sizeof(procinfo));
 
-    sattr.nLength = sizeof(sattr);
-    sattr.bInheritHandle = TRUE;
-    sattr.lpSecurityDescriptor = NULL;
-
-    if (!CreatePipe(&stdin_s, &stdin_m, &sattr, 0))
+    if (!CreatePipe(&stdin_s, &stdin_m, NULL, 0))
 	goto out_err_conv;
-    if (!SetHandleInformation(stdin_m, HANDLE_FLAG_INHERIT, 0))
+    if (!SetHandleInformation(stdin_s, HANDLE_FLAG_INHERIT,
+			      HANDLE_FLAG_INHERIT))
 	goto out_err_conv;
 
-    if (!CreatePipe(&stdout_m, &stdout_s, &sattr, 0))
+    if (!CreatePipe(&stdout_m, &stdout_s, NULL, 0))
 	goto out_err_conv;
-    if (!SetHandleInformation(stdout_m, HANDLE_FLAG_INHERIT, 0))
+    if (!SetHandleInformation(stdout_s, HANDLE_FLAG_INHERIT,
+			      HANDLE_FLAG_INHERIT))
 	goto out_err_conv;
 
     if (flags & GENSIO_EXEC_STDERR_TO_STDOUT) {
@@ -1264,9 +1305,7 @@ gensio_win_do_exec(struct gensio_os_funcs *o,
 			     0, TRUE, DUPLICATE_SAME_ACCESS))
 	    goto out_err_conv;
     } else if (rerr) {
-	if (!CreatePipe(&stderr_m, &stderr_s, &sattr, 0))
-	    goto out_err_conv;
-	if (!SetHandleInformation(stderr_m, HANDLE_FLAG_INHERIT, 0))
+	if (!CreatePipe(&stderr_m, &stderr_s, NULL, 0))
 	    goto out_err_conv;
     } else {
 	if (!DuplicateHandle(GetCurrentProcess(),
@@ -1290,7 +1329,7 @@ gensio_win_do_exec(struct gensio_os_funcs *o,
 		       cmdline,
 		       NULL,
 		       NULL,
-		       FALSE,
+		       TRUE,
 		       NORMAL_PRIORITY_CLASS,
 		       envb,
 		       start_dir,
@@ -1348,6 +1387,7 @@ read_token_info(HANDLE h, TOKEN_INFORMATION_CLASS type, void **rval,
 	/* This should fail. */
 	return ERROR_INVALID_DATA;
     err = GetLastError();
+    assert(err != 0); /* Keep scan-build happy, but probably a good idea. */
     if (err != ERROR_INSUFFICIENT_BUFFER)
 	return err;
     val = malloc(len);
@@ -1355,7 +1395,9 @@ read_token_info(HANDLE h, TOKEN_INFORMATION_CLASS type, void **rval,
 	return STATUS_NO_MEMORY;
     if (!GetTokenInformation(h, type, val, len, &len)) {
 	free(val);
-	return GetLastError();
+	err = GetLastError();
+	assert(err != 0); /* Keep scan-build happy, but probably a good idea. */
+	return err;
     }
     *rval = val;
     if (rlen)
@@ -1437,6 +1479,7 @@ get_logon_sid(SID *user, SID **logon_sid, bool *rfound)
     if (!WTSEnumerateSessionsA(WTS_CURRENT_SERVER_HANDLE, 0, 1,
 			       &sesinfo, &sescount)) {
 	err = GetLastError();
+	assert(err != 0); /* Keep scan-build happy, but probably a good idea. */
 	goto out_err;
     }
 
@@ -1944,16 +1987,17 @@ gensio_win_get_user_token(const char *user, const char *password,
  out_err:
     if (extra_groups)
 	free_groups(extra_groups);
-    if (htok)
-	CloseHandle(htok);
     if (usersid)
 	free(usersid);
     if (wuser)
 	free(wuser);
     if (logon_info)
 	free(logon_info);
-    if (profile_info.hProfile)
-	UnloadUserProfile(htok, profile_info.hProfile);
+    if (htok) {
+	CloseHandle(htok);
+	if (profile_info.hProfile)
+	  UnloadUserProfile(htok, profile_info.hProfile);
+    }
     if (profile)
 	LsaFreeReturnBuffer(profile);
     if (lsah)
@@ -2029,6 +2073,8 @@ gensio_win_pty_alloc(struct gensio_os_funcs *o,
  out_err_conv:
     err = gensio_os_err_to_err(o, GetLastError());
  out_err:
+    if (ptyh)
+	ClosePseudoConsole(ptyh);
     if (readh_m)
 	CloseHandle(readh_m);
     if (readh_s)
@@ -2037,8 +2083,6 @@ gensio_win_pty_alloc(struct gensio_os_funcs *o,
 	CloseHandle(writeh_m);
     if (writeh_s)
 	CloseHandle(writeh_s);
-    if (ptyh)
-	ClosePseudoConsole(ptyh);
     return err;
 }
 
@@ -2123,9 +2167,10 @@ start_pty_helper(struct gensio_os_funcs *o,
     UnmapViewOfFile(buf);
 
     /* Pass the handle in the command line. */
-    snprintf(cmd, sizeof(cmd), "gensio_pty_helper %lld %lu %lld",
-	     (intptr_t) shmem, (unsigned long) cmdline_len + 1,
-	     (intptr_t) ctl_s);
+    snprintf(cmd, sizeof(cmd), "gensio_pty_helper %llu %lu %llu",
+	     (unsigned long long) (intptr_t) shmem,
+	     (unsigned long) cmdline_len + 1,
+	     (unsigned long long) (intptr_t) ctl_s);
 
     if (!CreateProcessAsUserA(tokh,
 			      NULL,
@@ -2166,13 +2211,14 @@ gensio_win_pty_start(struct gensio_os_funcs *o,
     char *cmdline, *envb = NULL;
     STARTUPINFOEX si;
     PROCESS_INFORMATION procinfo;
-    size_t len;
+    SIZE_T len;
     HANDLE tokh = NULL, imptokh = NULL, control = NULL;
     int err;
     bool setuser = true;
     bool attrs_added = false;
 
     memset(&si, 0, sizeof(si));
+    memset(&procinfo, 0, sizeof(procinfo)); /* Keep scan-build happy. */
 
     err = argv_to_win_cmdline(o, argv, &cmdline);
     if (err)
@@ -2405,8 +2451,8 @@ struct gensio_unix_termios {
     g_termios curr_termios;
     bool break_set;
 #if HAVE_DECL_TIOCSRS485
-    bool rs485_applied;
     struct serial_rs485 rs485;
+    struct serial_rs485 orig_rs485;
 #endif
 };
 
@@ -2543,6 +2589,9 @@ gensio_unix_setup_termios(struct gensio_os_funcs *o, int fd,
 
     t->orig_termios = t->curr_termios;
     ioctl(fd, TIOCMGET, &t->orig_mctl);
+#if HAVE_DECL_TIOCSRS485
+    ioctl(fd, TIOCGRS485, &t->orig_rs485);
+#endif
 
     s_cfmakeraw(&t->curr_termios);
 
@@ -2563,6 +2612,9 @@ gensio_unix_cleanup_termios(struct gensio_os_funcs *o,
 {
     if (!*it)
 	return;
+#if HAVE_DECL_TIOCSRS485
+    ioctl(fd, TIOCSRS485, &(*it)->orig_rs485);
+#endif
     ioctl(fd, TIOCMSET, &(*it)->orig_mctl);
     set_termios(fd, &(*it)->orig_termios);
     o->free(o, *it);
@@ -2948,10 +3000,19 @@ gensio_unix_termios_control(struct gensio_os_funcs *o, int op, bool get,
 	if (get) {
 	    *((int *) val) = !!(t->curr_termios.c_cflag & HUPCL);
 	} else {
-	    if (val)
+	    /*
+	     * Setup hupcl on the original termios that will be
+	     * restored as wellas the current termios.  Otherwise when
+	     * we close will will restore the original termios and
+	     * hupcl won't work.
+	     */
+	    if (val) {
+		t->orig_termios.c_cflag |= HUPCL;
 		t->curr_termios.c_cflag |= HUPCL;
-	    else
+	    } else {
+		t->orig_termios.c_cflag &= ~HUPCL;
 		t->curr_termios.c_cflag &= ~HUPCL;
+	    }
 	}
 	break;
 
@@ -2981,15 +3042,12 @@ gensio_unix_termios_control(struct gensio_os_funcs *o, int op, bool get,
 	if (rv) {
 	    rv = gensio_os_err_to_err(o, errno);
 #if HAVE_DECL_TIOCSRS485
+	} else if (t->rs485.flags & SER_RS485_ENABLED) {
+	    if (ioctl(fd, TIOCSRS485, &t->rs485) < 0)
+		rv = gensio_os_err_to_err(o, errno);
 	} else {
-	    bool enabled = !!(t->rs485.flags & SER_RS485_ENABLED);
-
-	    if (enabled != t->rs485_applied) {
-		if (ioctl(fd, TIOCSRS485, &t->rs485) < 0)
-		    rv = gensio_os_err_to_err(o, errno);
-		else
-		    t->rs485_applied = enabled;
-	    }
+	    /* If the disable fails, we don't care. */
+	    ioctl(fd, TIOCSRS485, &t->rs485);
 #endif
 	}
 	break;
@@ -3061,13 +3119,13 @@ gensio_unix_termios_control(struct gensio_os_funcs *o, int op, bool get,
 	}
 	modemstate = 0;
 	if (nval & TIOCM_CD)
-	    modemstate |= SERGENSIO_MODEMSTATE_CD;
+	    modemstate |= GENSIO_SER_MODEMSTATE_CD;
 	if (nval & TIOCM_RI)
-	    modemstate |= SERGENSIO_MODEMSTATE_RI;
+	    modemstate |= GENSIO_SER_MODEMSTATE_RI;
 	if (nval & TIOCM_DSR)
-	    modemstate |= SERGENSIO_MODEMSTATE_DSR;
+	    modemstate |= GENSIO_SER_MODEMSTATE_DSR;
 	if (nval & TIOCM_CTS)
-	    modemstate |= SERGENSIO_MODEMSTATE_CTS;
+	    modemstate |= GENSIO_SER_MODEMSTATE_CTS;
 	*((int *) val) = modemstate;
 	break;
 
@@ -3246,18 +3304,36 @@ gensio_unix_do_exec(struct gensio_os_funcs *o,
 }
 
 int
-gensio_unix_pty_alloc(struct gensio_os_funcs *o, int *rfd)
+gensio_unix_pty_alloc(struct gensio_os_funcs *o, int *rfd, int *rsfd)
 {
-    int fd = posix_openpt(O_RDWR | O_NOCTTY);
+#if USE_OPENPTY
+    int fd;
+    int sfd = -1;
+
+    if (openpty(&fd, &sfd, NULL, NULL, NULL) == -1)
+	return gensio_os_err_to_err(o, errno);
+#else
+    int fd = posix_openpt(O_RDWR | O_NOCTTY), sfd = -1;
     if (fd == -1)
 	return gensio_os_err_to_err(o, errno);
+#endif
+
+    /* Set the owner of the slave PT. */
+    if (grantpt(fd) < 0) {
+	close(fd);
+	if (sfd != -1)
+	    close(sfd);
+	return gensio_os_err_to_err(o, errno);
+    };
+
     *rfd = fd;
+    *rsfd = sfd;
     return 0;
 }
 
 int
 gensio_unix_pty_start(struct gensio_os_funcs *o,
-		      int pfd, const char **argv, const char **env,
+		      int pfd, int *psfd, const char **argv, const char **env,
 		      const char *start_dir, pid_t *rpid)
 {
     const char *pgm;
@@ -3283,10 +3359,6 @@ gensio_unix_pty_start(struct gensio_os_funcs *o,
 	int i, openfiles = sysconf(_SC_OPEN_MAX);
 	int fd;
 
-	/* Set the owner of the slave PT. */
-	if (grantpt(pfd) < 0)
-	    exit(1);
-
 	if (start_dir) {
 	    if (chdir(start_dir)) {
 		fprintf(stderr, "pty fork: chdir to %s failed: %s",
@@ -3301,11 +3373,15 @@ gensio_unix_pty_start(struct gensio_os_funcs *o,
 	    exit(1);
 	}
 
-	fd = open(slave, O_RDWR);
-	if (fd == -1) {
-	    fprintf(stderr, "pty fork: failed to open slave terminal: %s\r\n",
-		    strerror(errno));
-	    exit(1);
+	if (*psfd == -1) {
+	    fd = open(slave, O_RDWR);
+	    if (fd == -1) {
+		fprintf(stderr, "pty fork: failed to open slave terminal: %s\r\n",
+			strerror(errno));
+		exit(1);
+	    }
+	} else {
+	    fd = *psfd;
 	}
 
 #if defined(TIOCSCTTY) && !defined(linux)
@@ -3361,6 +3437,10 @@ gensio_unix_pty_start(struct gensio_os_funcs *o,
 		strerror(errno));
 	exit(1); /* Only reached on error. */
     }
+    if (*psfd != -1) {
+	close(*psfd);
+	*psfd = -1;
+    }
  skip_child:
     *rpid = pid;
     return 0;
@@ -3479,10 +3559,6 @@ gensio_os_loadlib(struct gensio_os_funcs *o, const char *name)
 
 #ifdef ENABLE_INTERNAL_TRACE
 
-#include <pthread_handler.h>
-
-#define TRACEBACK_DEPTH 1
-
 #define MEM_MAGIC 0xddf0983aec9320b0
 #define MEM_BUFFER 32
 struct mem_header {
@@ -3617,7 +3693,8 @@ gensio_memtrack_cleanup(struct gensio_memtrack *m)
 }
 
 void *
-gensio_i_zalloc(struct gensio_memtrack *m, unsigned int size)
+gensio_i_zalloc(struct gensio_memtrack *m, unsigned int size,
+		void *caller[], unsigned int caller_size)
 {
     unsigned char *b;
 
@@ -3636,20 +3713,12 @@ gensio_i_zalloc(struct gensio_memtrack *m, unsigned int size)
 	h->magic = MEM_MAGIC;
 	h->freed = 0;
 	h->size = size;
-#if _MSC_VER
-	h->alloc_bt[0] = _ReturnAddress();
-#else
-	h->alloc_bt[0] = __builtin_return_address(0);
-#if TRACEBACK_DEPTH > 1
-	h->alloc_bt[1] = __builtin_return_address(1);
-#if TRACEBACK_DEPTH > 2
-	h->alloc_bt[2] = __builtin_return_address(2);
-#if TRACEBACK_DEPTH > 3
-	h->alloc_bt[3] = __builtin_return_address(3);
-#endif
-#endif
-#endif
-#endif
+	if (caller) {
+	    unsigned int i;
+
+	    for (i = 0; i < caller_size && i < 4; i++)
+		h->alloc_bt[i] = caller[i];
+	}
 	b += sizeof(struct mem_header);
 	mem_fill(h->filler);
 	mem_fill(b + size);
@@ -3665,7 +3734,8 @@ gensio_i_zalloc(struct gensio_memtrack *m, unsigned int size)
 }
 
 void
-gensio_i_free(struct gensio_memtrack *m, void *data)
+gensio_i_free(struct gensio_memtrack *m, void *data,
+	      void *caller[], unsigned int caller_size)
 {
     if (m) {
 	unsigned char *b = data;
@@ -3673,6 +3743,12 @@ gensio_i_free(struct gensio_memtrack *m, void *data)
 	struct gensio_link *l;
 	bool err;
 
+	if (caller) {
+	    unsigned int i;
+
+	    for (i = 0; i < caller_size && i < 4; i++)
+		h->free_bt[0] = caller[i];
+	}
 #if _MSC_VER
 	h->free_bt[0] = _ReturnAddress();
 #else
@@ -3733,7 +3809,8 @@ gensio_memtrack_cleanup(struct gensio_memtrack *m)
 }
 
 void *
-gensio_i_zalloc(struct gensio_memtrack *m, unsigned int size)
+gensio_i_zalloc(struct gensio_memtrack *m, unsigned int size,
+		void *caller[], unsigned int caller_size)
 {
     void *d = malloc(size);
     if (d)
@@ -3742,7 +3819,8 @@ gensio_i_zalloc(struct gensio_memtrack *m, unsigned int size)
 }
 
 void
-gensio_i_free(struct gensio_memtrack *m, void *data)
+gensio_i_free(struct gensio_memtrack *m, void *data,
+	      void *caller[], unsigned int caller_size)
 {
     free(data);
 }
@@ -3953,3 +4031,74 @@ gensio_os_funcs_get_data(struct gensio_os_funcs *o)
 {
     return o->other_data;
 }
+
+/*
+ * The once operation is a nasty piece of code.
+ *
+ * This basically implements a recursive lock on the once operation.
+ * There are unfortunate places one a once calls another once and
+ * where these can actually recurse if you don't stop them.
+ *
+ * once_lock protects once_next and once_owner.  once_lock2 is used to
+ * let only one once operation in at a time and to protect
+ * once->called.  You can claim once_lock if holding once_lock2, but
+ * not the other way around.
+ *
+ * The called field in the once is used to tell if we have called it.
+ * It will be 2 if it is in the once function (to detect recursive
+ * calls) and 1 if the once operation has been completed.
+ */
+static lock_type once_lock = LOCK_INITIALIZER;
+static lock_type once_lock2 = LOCK_INITIALIZER;
+static unsigned int once_nest;
+static lock_thread_type once_owner;
+
+void
+gensio_call_once(struct gensio_os_funcs *f, struct gensio_once *once,
+		 void (*func)(void *cb_data), void *cb_data)
+{
+    lock_thread_type tid;
+    bool do_unlock = true;
+
+    if (once->called == 1)
+	/* once has been done, just return. */
+	return;
+
+    LOCK(&once_lock);
+    tid = LOCK_GET_THREAD_ID;
+    if (once_nest > 0 && LOCK_THREAD_ID_EQUAL(tid, once_owner)) {
+	/* Called recursively. */
+	once_nest++;
+	UNLOCK(&once_lock);
+	do_unlock = false;
+    } else {
+	/*
+	 * First time into the once for this thread, claim the that
+	 * only allows one caller.
+	 */
+	UNLOCK(&once_lock);
+	LOCK(&once_lock2);
+
+	LOCK(&once_lock);
+	once_owner = tid;
+	once_nest++;
+	UNLOCK(&once_lock);
+    }
+
+    if (!once->called) {
+	/*
+	 * Mark that we are in the call.  This way if we recurse into
+	 * the same once, we know to not call it again.  Don't set it
+	 * to 1 until we are finished processing.
+	 */
+	once->called = 2;
+	func(cb_data);
+	once->called = 1;
+    }
+
+    LOCK(&once_lock);
+    once_nest--;
+    UNLOCK(&once_lock);
+    if (do_unlock)
+	UNLOCK(&once_lock2);
+}
