@@ -78,7 +78,8 @@ struct pty_data {
     bool exit_code_set;
 };
 
-static int pty_check_open(void *handler_data, struct gensio_iod *iod)
+static int pty_check_open(void *handler_data, struct gensio_iod *iod,
+			  gensio_time *timeout)
 {
     return 0;
 }
@@ -307,7 +308,8 @@ gensio_setup_child_on_pty(struct pty_data *tdata)
 }
 
 static int
-pty_sub_open(void *handler_data, struct gensio_iod **riod)
+pty_sub_open(void *handler_data, struct gensio_iod **riod,
+	     gensio_time *timeout)
 {
     struct pty_data *tdata = handler_data;
     int err;
@@ -363,8 +365,10 @@ pty_check_close(void *handler_data, struct gensio_iod *iod,
     err = pty_check_exit_code(tdata);
     if (err == GE_INPROGRESS) {
 	/* FIXME - this should probably be configurable. */
-	if (tdata->check_close_count >= 500) /* Wait for 5 seconds. */
+	if (tdata->check_close_count >= 500) { /* Wait for 5 seconds. */
+	    err = GE_TIMEDOUT;
 	    goto out_finish;
+	}
 	tdata->check_close_count++;
 	timeout->secs = 0;
 	timeout->nsecs = 10000000;
